"""Test context plugin."""
from .. import util


class TestContext(util.PluginTestCase):
    """Test context plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: context
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.context:
                  context_visible_first: true
                  escapes: '\\[\\`]'
                  delimiters:
                  # Ignore multiline content between fences (fences can have 3 or more back ticks)
                  # ```
                  # content
                  # ```
                  - open: '(?s)^(?P<open> *`{{3,}})$'
                    close: '^(?P=open)$'
                  # Ignore text between inline back ticks
                  - open: '(?P<open>`+)'
                    close: '(?P=open)'
            """
        ).format(self.tempdir)
        self.mktemp('.context.yml', config, 'utf-8')

    def test_context(self):
        """Test text."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            {}
            ```
            djlask dkjsl dksj
            kudk alks
            ```

            Line with `incoretc txet`.
            """
        ).format('\n'.join(bad_words + good_words))
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.context.yml', bad_words)


class TestContextChained(util.PluginTestCase):
    """Test chained context plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: context
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.context:
                  context_visible_first: true
                  escapes: '\\[\\`]'
                  delimiters:
                  # Ignore multiline content between fences (fences can have 3 or more back ticks)
                  # ```
                  # content
                  # ```
                  - open: '(?s)^(?P<open> *`{{3,}})$'
                    close: '^(?P=open)$'
                  # Ignore text between inline back ticks
                  - open: '(?P<open>`+)'
                    close: '(?P=open)'
            """
        ).format(self.tempdir)
        self.mktemp('.context.yml', config, 'utf-8')

    def test_context_after_text(self):
        """Test text."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        template = self.dedent(
            """
            {}
            ```
            djlask dkjsl dksj
            kudk alks
            ```

            Line with `incoretc txet`.
            """
        ).format('\n'.join(bad_words + good_words))
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.context.yml', bad_words)



"""
HTML filter.

Detect encoding from HTML header.
"""
import re
import codecs
import soupsieve as sv
from . import xml
from collections import deque

RE_HTML_ENCODE = re.compile(
    br'''(?xi)
    <\s*meta(?!\s*(?:name|value)\s*=)(?:[^>]*?content\s*=[\s"']*)?(?:[^>]*?)[\s"';]*charset\s*=[\s"']*([^\s"'/>]*)
    '''
)

MODE = {'html': 'lxml', 'xhtml': 'xml', 'html5': 'html5lib'}


class HtmlFilter(xml.XmlFilter):
    """Spelling Python."""

    break_tags = {
        # Block level elements (and other blockish elements)
        'address', 'article', 'aside', 'blockquote', 'details', 'dialog', 'dd',
        'div', 'dl', 'dt'
        'fieldset', 'figcaption', 'figure', 'footer', 'form', 'h1', 'h2', 'h3',
        'h4', 'h5', 'h6', 'header', 'hgroup', 'hr', 'li', 'main', 'menu', 'nav', 'ol', 'p', 'pre',
        'section', 'table', 'ul',
        'canvas', 'group', 'iframe', 'math', 'noscript', 'output',
        'script', 'style', 'table', 'video', 'body', 'head'
    }

    default_capture = ['*|*:not(script, style)']

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "comments": True,
            "mode": "html",
            "attributes": [],
            "break_tags": [],
            "ignores": [],
            "captures": self.default_capture,
            "namespaces": {}
        }

    def validate_options(self, k, v):
        """Validate options."""

        super().validate_options(k, v)
        if k == 'mode' and v not in MODE:
            raise ValueError("{}: '{}' is not a valid value for '{}'".format(self.__class__.__name, v, k))

    def setup(self):
        """Setup."""

        self.user_break_tags = set(self.config['break_tags'])
        self.comments = self.config.get('comments', True) is True
        self.attributes = set(self.config['attributes'])
        self.type = self.config['mode']
        if self.type not in MODE:
            self.type = 'html'
        self.parser = MODE[self.type]
        ignores = ','.join(self.config['ignores'])
        self.ignores = sv.compile(ignores, self.config['namespaces']) if ignores.strip() else None
        captures = ','.join(self.config['captures'])
        self.captures = sv.compile(captures, self.config['namespaces']) if captures.strip() else None

    def header_check(self, content):
        """Special HTML encoding check."""

        encode = None

        # Look for meta charset
        m = RE_HTML_ENCODE.search(content)
        if m:
            enc = m.group(1).decode('ascii')

            try:
                codecs.getencoder(enc)
                encode = enc
            except LookupError:
                pass
        else:
            encode = self._has_xml_encode(content)
        return encode

    def is_break_tag(self, el):
        """Check if tag is an element we should break on."""

        name = sv.util.lower(el.name) if self.type != 'xhtml' else el.name
        return name in self.break_tags or name in self.user_break_tags

    def get_classes(self, el):
        """Get classes."""

        if self.type != 'xhtml':
            return el.attrs.get('class', [])
        else:
            return [c for c in el.attrs.get('class', '').strip().split(' ') if c]

    def format_blocks(self):
        """Format the text as for a block."""

        block_text = []
        for el, text in self._block_text.items():
            content = ''.join(text)
            if content:
                block_text.append((content, self.construct_selector(el)))
        return block_text

    def construct_selector(self, el, attr=''):
        """Construct an selector for context."""

        selector = deque()
        ancestor = el
        while ancestor and ancestor.parent:
            if ancestor is not el:
                selector.appendleft(ancestor.name)
            else:
                tag = ancestor.name
                prefix = ancestor.prefix
                classes = self.get_classes(ancestor)
                tag_id = ancestor.attrs.get('id', '').strip()
                sel = ''
                if prefix:
                    sel += prefix + '|'
                sel += tag
                if tag_id:
                    sel += '#' + tag_id
                if classes:
                    sel += '.' + '.'.join(classes)
                if attr:
                    sel += '[%s]' % attr
                selector.appendleft(sel)
            ancestor = ancestor.parent
        return '>'.join(selector)


def get_plugin():
    """Return the filter."""

    return HtmlFilter



"""Test text plugin."""
from . import util
from wcmatch._wcparse import PatternLimitException


class TestGlob(util.PluginTestCase):
    """Test `glob` behavior."""

    def test_glob_limit(self):
        """Test `glob` with a custom limit."""

        config = self.dedent(
            """
            matrix:
            - name: glob
              default_encoding: utf-8
              glob_pattern_limit: 10
              sources:
              - '{}/**/test-{{1..11}}.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline: null
            """
        ).format(self.tempdir)
        self.mktemp('.glob.yml', config, 'utf-8')
        with self.assertRaises(PatternLimitException):
            self.assert_spellcheck('.glob.yml', [])

    def test_glob_no_limit(self):
        """Test when there is no limit for `glob` expansion patterns."""

        config = self.dedent(
            """
            matrix:
            - name: glob
              default_encoding: utf-8
              glob_pattern_limit: 0
              sources:
              - '{}/**/test-{{1..11}}.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline: null
            """
        ).format(self.tempdir)
        self.mktemp('.glob.yml', config, 'utf-8')

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        self.mktemp('test-1.txt', '\n'.join(bad_words + good_words), 'utf-8')
        self.assert_spellcheck('.glob.yml', bad_words)


class TestNoPipeline(util.PluginTestCase):
    """Test no pipeline."""

    def test_no_pipeline(self):
        """Test with no pipeline."""

        config = self.dedent(
            """
            matrix:
            - name: no_pipeline
              default_encoding: utf-8
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline: null
            """
        ).format(self.tempdir)
        self.mktemp('.nopipeline.yml', config, 'utf-8')

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        self.mktemp('test.txt', '\n'.join(bad_words + good_words), 'utf-8')
        self.assert_spellcheck('.nopipeline.yml', bad_words)

    def test_no_pipeline_wordlist(self):
        """Test text."""

        config = self.dedent(
            """
            matrix:
            - name: no_pipeline
              default_encoding: utf-8
              sources:
              - '{temp}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline: null
              dictionary:
                wordlists:
                - '{temp}/mydict.wl'
                output: '{temp}/mydict.dic'
            """
        ).format(temp=self.tempdir)
        self.mktemp('.nopipeline.yml', config, 'utf-8')

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']
        self.mktemp('mydict.wl', '\n'.join(bad_words), 'utf-8')
        self.mktemp('test.txt', '\n'.join(bad_words + good_words), 'utf-8')
        self.assert_spellcheck('.nopipeline.yml', [])


class TestNameGroup(util.PluginTestCase):
    """Test name group behavior."""

    def setup_fs(self):
        """Setup files."""

        good_words = ['yes', 'word']
        self.bad_words1 = ['helo', 'begn']
        self.bad_words2 = ['oleh', 'ngeb']
        self.bad_words3 = ['eolh', 'ngbe']
        self.mktemp('test1.txt', '\n'.join(self.bad_words1 + good_words), 'utf-8')
        self.mktemp('test2.txt', '\n'.join(self.bad_words2 + good_words), 'utf-8')
        self.mktemp('test3.txt', '\n'.join(self.bad_words3 + good_words), 'utf-8')

        config = self.dedent(
            """
            matrix:
            - name: name
              group: group1
              default_encoding: utf-8
              sources:
              - '{temp}/**/test1.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
            # Purposely reuse same name
            - name: name
              group: group2
              default_encoding: utf-8
              sources:
              - '{temp}/**/test2.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
            - name: name3
              group: group1
              default_encoding: utf-8
              sources:
              - '{temp}/**/test3.txt'
              - '{temp}/**/test4.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
            """
        ).format(temp=self.tempdir)
        self.mktemp('.source.yml', config, 'utf-8')

    def test_all(self):
        """Test all."""

        self.assert_spellcheck('.source.yml', self.bad_words1 + self.bad_words2 + self.bad_words3)

    def test_name(self):
        """Test name works."""

        self.assert_spellcheck('.source.yml', self.bad_words2, names=['name'])

    def test_group(self):
        """Test group works."""

        self.assert_spellcheck('.source.yml', self.bad_words1 + self.bad_words3, groups=['group1'])

    def test_no_sources(self):
        """Test missing source match raises a `ValueError`."""

        config = self.dedent(
            """
            matrix:
            - name: name
              group: group1
              default_encoding: utf-8
              sources:
              - '{temp}/**/test4.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
            """
        ).format(temp=self.tempdir)
        self.mktemp('.nosource.yml', config, 'utf-8')
        with self.assertRaises(RuntimeError) as excctxt:
            self.assert_spellcheck('.nosource.yml', [])
        self.assertIn('test4.txt', str(excctxt.exception))

    def test_sources(self):
        """Test source override."""

        self.assert_spellcheck('.source.yml', self.bad_words1, names=['name'], sources=[self.tempdir + '/**/test1.txt'])

    def test_sources_invalid(self):
        """Test that source override won't work if multiple names are defined."""

        self.assert_spellcheck(
            '.source.yml', self.bad_words2, names=['name', 'other_name'], sources=[self.tempdir + '/**/test1.txt']
        )

    def test_no_matrix(self):
        """Test a `ValueError` is raised if the configuration has no matrix."""

        config = self.dedent(
            """
            something: []
            """
        )

        self.mktemp('.source.yml', config, 'utf-8')
        with self.assertRaises(KeyError):
            self.assert_spellcheck('.source.yml', [])

    def test_bad_name(self):
        """Test bad name."""

        config = self.dedent(
            """
            matrix:
            - name: other
            """
        )
        self.mktemp('.source.yml', config, 'utf-8')
        with self.assertRaises(ValueError):
            self.assert_spellcheck('.source.yml', [], names=['name'])

    def test_no_name(self):
        """Test no name."""

        config = self.dedent(
            """
            matrix: []
            """
        )
        self.mktemp('.source.yml', config, 'utf-8')
        with self.assertRaises(ValueError):
            self.assert_spellcheck('.source.yml', [])



"""Main."""
import sys
import argparse
from pyspelling import spellcheck, __version__


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='spellcheck', description='Spell checking tool.')
    # Flag arguments
    parser.add_argument('--version', action='version', version=('%(prog)s ' + __version__))
    parser.add_argument('--debug', action='store_true', default=False, help=argparse.SUPPRESS)
    parser.add_argument('--verbose', '-v', action='count', default=0, help="Verbosity level.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--name', '-n', action='append', help="Specific spelling task by name to run.")
    group.add_argument('--group', '-g', action='append', help="Specific spelling task groun to run.")
    parser.add_argument('--binary', '-b', action='store', default='', help="Provide path to spell checker's binary.")
    parser.add_argument('--config', '-c', action='store', default='', help="Spelling config.")
    parser.add_argument('--source', '-S', action='append', help="Specify override file pattern.")
    parser.add_argument(
        '--spellchecker', '-s', action='store', default='', help="Choose between aspell and hunspell"
    )
    args = parser.parse_args()

    return run(
        args.config,
        names=args.name,
        groups=args.group,
        binary=args.binary,
        spellchecker=args.spellchecker,
        sources=args.source,
        verbose=args.verbose,
        debug=args.debug
    )


def run(config, **kwargs):
    """Run."""

    names = kwargs.get('names', [])
    groups = kwargs.get('groups', [])
    binary = kwargs.get('binary', '')
    spellchecker = kwargs.get('spellchecker', '')
    verbose = kwargs.get('verbose', 0)
    sources = kwargs.get('sources', [])
    debug = kwargs.get('debug', False)

    fail = False
    count = 0
    for results in spellcheck(
        config,
        names=names,
        groups=groups,
        binary=binary,
        checker=spellchecker,
        sources=sources,
        verbose=verbose,
        debug=debug
    ):
        count += 1
        if results.error:
            fail = True
            print('ERROR: %s -- %s' % (results.context, results.error))
        elif results.words:
            fail = True
            print('Misspelled words:\n<%s> %s' % (results.category, results.context))
            print('-' * 80)
            for word in results.words:
                print(word)
            print('-' * 80)
            print('')

    if fail:
        print('!!!Spelling check failed!!!')
    else:
        print('Spelling check passed :)')

    return fail


if __name__ == "__main__":
    sys.exit(main())



"""Filter Plugins."""
import re
import codecs
import contextlib
import mmap
import os
from collections import namedtuple
from .. import plugin
from .. import util

PYTHON_ENCODING_NAMES = {
    'iso-8859-8-i': 'iso-8859-8',
    'macintosh': 'mac-roman',
    'windows-874': 'cp874',
    'x-mac-cyrillic': 'mac-cyrillic',
    "x-sjis": "shift-jis"
}

RE_UTF_BOM = re.compile(
    b'^(?:(' +
    codecs.BOM_UTF8 +
    b')[\x00-\xFF]{,2}|(' +
    codecs.BOM_UTF32_BE +
    b')|(' +
    codecs.BOM_UTF32_LE +
    b')|(' +
    codecs.BOM_UTF16_BE +
    b')|(' +
    codecs.BOM_UTF16_LE +
    b'))'
)

RE_CATEGORY_NAME = re.compile(r'^[-a-z0-9_]+$', re.I)

RE_LINE_ENDINGS = re.compile(r'(?:\r\n|(?!\r\n)[\n\r])')

BINARY_ENCODE = ''


class SourceText(namedtuple('SourceText', ['text', 'context', 'encoding', 'category', 'error'])):
    """Source text."""

    __slots__ = ()

    def __new__(cls, text, context, encoding, category, error=None):
        """Allow defaults."""

        encoding = PYTHON_ENCODING_NAMES.get(encoding, encoding).lower()

        if encoding == 'utf-8-sig':
            encoding = 'utf-8'
        if encoding.startswith('utf-16'):
            encoding = 'utf-16'
        elif encoding.startswith('utf-32'):
            encoding = 'utf-32'

        if encoding:
            encoding = codecs.lookup(encoding).name

        if RE_CATEGORY_NAME.match(category) is None and error is None:
            raise ValueError('Invalid category name in SourceText!')

        return super().__new__(cls, text, context, encoding, category, error)

    def _is_bytes(self):
        """Is bytes."""

        return isinstance(self.text, bytes)

    def _has_error(self):
        """Check if object has an error associated with it."""

        return self.error is not None


class Filter(plugin.Plugin):
    """Spelling language."""

    MAX_GUESS_SIZE = 31457280
    CHECK_BOM = True

    def __init__(self, options, default_encoding='utf-8'):
        """Initialize."""

        self.default_encoding = PYTHON_ENCODING_NAMES.get(default_encoding, default_encoding).lower()
        super().__init__(options)
        self.setup()

    def _is_very_large(self, size):
        """Check if content is very large."""

        return size >= self.MAX_GUESS_SIZE

    def _verify_encoding(self, enc):
        """Verify encoding is okay."""

        enc = PYTHON_ENCODING_NAMES.get(enc, enc)
        try:
            codecs.getencoder(enc)
            encoding = enc
        except LookupError:
            encoding = None
        return encoding

    def has_bom(self, f):
        """Check for UTF8, UTF16, and UTF32 BOMs."""

        content = f.read(4)
        encoding = None
        m = RE_UTF_BOM.match(content)
        if m is not None:
            if m.group(1):
                encoding = 'utf-8-sig'
            elif m.group(2):
                encoding = 'utf-32'
            elif m.group(3):
                encoding = 'utf-32'
            elif m.group(4):
                encoding = 'utf-16'
            elif m.group(5):
                encoding = 'utf-16'
        return encoding

    def _utf_strip_bom(self, encoding):
        """Return an encoding that will ignore the BOM."""

        if encoding is None:
            pass
        elif encoding.lower() == 'utf-8':
            encoding = 'utf-8-sig'
        elif encoding.lower().startswith('utf-16'):
            encoding = 'utf-16'
        elif encoding.lower().startswith('utf-32'):
            encoding = 'utf-32'
        return encoding

    def _detect_buffer_encoding(self, f):
        """Guess by checking BOM, and checking `_special_encode_check`, and using memory map."""

        encoding = None
        with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
            encoding = self._analyze_file(m)
        return encoding

    def _analyze_file(self, f):
        """Analyze the file."""

        f.seek(0)
        # Check for BOMs
        if self.CHECK_BOM:
            encoding = self.has_bom(f)
            f.seek(0)
        else:
            util.warn_deprecated(
                "'CHECK_BOM' attribute is deprecated. "
                "Please override 'has_bom` function to control or avoid BOM detection."
            )
        # Check file extensions
        if encoding is None:
            encoding = self._utf_strip_bom(self.header_check(f.read(1024)))
            f.seek(0)
        if encoding is None:
            encoding = self._utf_strip_bom(self.content_check(f))
            f.seek(0)

        return encoding

    def _detect_encoding(self, source_file):
        """Detect encoding."""

        encoding = self._guess(source_file)
        # If we didn't explicitly detect an encoding, assume default.
        if encoding is None:
            encoding = self.default_encoding

        return encoding

    def _run_first(self, source_file):
        """Run on as first in chain."""

        self.reset()
        self.current_encoding = self.default_encoding
        encoding = None
        try:
            encoding = self._detect_encoding(source_file)
            content = self.filter(source_file, encoding)
        except UnicodeDecodeError:
            if not encoding or encoding != self.default_encoding:
                content = self.filter(source_file, self.default_encoding)
            else:
                raise
        return content

    def _run(self, source):
        """Run chained."""

        self.reset()
        return self.sfilter(source)

    def _guess(self, filename):
        """Guess the encoding and decode the content of the file."""

        encoding = None

        file_size = os.path.getsize(filename)
        # If the file is really big, lets just call it binary.
        # We don't have time to let Python chug through a massive file.
        if not self._is_very_large(file_size):
            with open(filename, "rb") as f:
                if file_size == 0:
                    encoding = 'ascii'
                else:
                    encoding = self._detect_buffer_encoding(f)
                    if encoding is None:
                        raise UnicodeDecodeError('None', b'', 0, 0, 'Unicode cannot be detected.')
                    if encoding != BINARY_ENCODE:
                        encoding = self._verify_encoding(encoding)
        else:  # pragma: no cover
            raise UnicodeDecodeError('None', b'', 0, 0, 'Unicode detection is not applied to very large files!')

        return encoding

    def norm_nl(self, text):
        """Normalize line endings."""

        return RE_LINE_ENDINGS.sub('\n', text)

    def content_check(self, f):
        """File content check."""

        return None

    def header_check(self, content):
        """Special encode check."""

        return None

    def filter(self, source_file, encoding):  # noqa A001
        """Open and filter the file from disk."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        return [SourceText(text, source_file, encoding, 'text')]

    def sfilter(self, source):
        """Execute filter."""

        return [SourceText(source.text, source.context, source.encoding, 'text')]



"""Flow control."""
from .. import plugin

ALLOW = 0
SKIP = 1
HALT = 2


class FlowControl(plugin.Plugin):
    """Control flow of objects in the pipeline."""

    def __init__(self, options):
        """Initialization."""

        super().__init__(options)
        self.setup()

    def _run(self, category):
        """Run as chained plugin."""

        return self.adjust_flow(category)

    def adjust_flow(self, category):
        """Adjust the flow of source control objects."""

        return ALLOW



"""
Open Document Format (ODF) filter.

Parse the `text` namespace in `content.xml` in ODF zip files.
"""
import zipfile
import io
import bs4
import codecs
from .. import filters
from . import xml
from wcmatch import glob
import soupsieve as sv

MIMEMAP = {
    'application/vnd.oasis.opendocument.spreadsheet': 'ods',
    'application/vnd.oasis.opendocument.presentation': 'odp',
    'application/vnd.oasis.opendocument.text': 'odt'
}


class OdfFilter(xml.XmlFilter):
    """Spelling Python."""

    FLAGS = glob.G | glob.N | glob.B | glob.S | glob.O

    default_capture = ['text|*']

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {}

    def setup(self):
        """Setup."""

        self.additional_context = ''
        self.comments = False
        self.attributes = []
        self.parser = 'xml'
        self.type = None
        self.filepattern = 'content.xml'
        self.namespaces = {
            'text': 'urn:oasis:names:tc:opendocument:xmlns:text:1.0',
            'draw': 'urn:oasis:names:tc:opendocument:xmlns:drawing:1.0'
        }
        self.ignores = None
        self.captures = sv.compile(','.join(self.default_capture), self.namespaces)

    def has_bom(self, filestream):
        """Check if has BOM."""

        content = filestream.read(4)
        if content == b'PK\x03\x04':
            # Zip file found.
            # Return `BINARY_ENCODE` as content is binary type,
            # but don't return None which means we don't know what we have.
            return filters.BINARY_ENCODE
        # Not a zip file, so maybe a flat ODF XML file.
        return super().has_bom(filestream)

    def determine_file_type(self, z):
        """Determine file type."""

        mimetype = z.read('mimetype').decode('utf-8').strip()
        self.type = MIMEMAP[mimetype]

    def get_zip_content(self, filename):
        """Get zip content."""

        with zipfile.ZipFile(filename, 'r') as z:
            self.determine_file_type(z)
            for item in z.infolist():
                if glob.globmatch(item.filename, self.filepattern, flags=self.FLAGS):
                    yield z.read(item.filename), item.filename

    def get_content(self, zipbundle):
        """Get content."""

        for content, filename in self.get_zip_content(zipbundle):
            with io.BytesIO(content) as b:
                encoding = self._analyze_file(b)
                if encoding is None:
                    encoding = self.default_encoding
                b.seek(0)
                text = b.read().decode(encoding)
            yield text, filename, encoding

    def is_break_tag(self, el):
        """Break on specified boundaries."""

        should_break = False
        if self.type == 'odp':
            if el.name == 'page' and el.namespace and el.namespace == self.namespaces['draw']:
                should_break = True
        return should_break

    def soft_break(self, el, text):
        """Apply soft break if needed."""

        if el.name == 'p' and el.namespace and el.namespace == self.namespaces["text"]:
            text.append('\n')

    def format_blocks(self):
        """Format the text as for a block."""

        block_text = []
        for el, text in self._block_text.items():
            self.soft_break(el, text)
            content = ''.join(text)
            if content:
                block_text.append((content, self.additional_context + self.construct_selector(el)))
        return block_text

    def extract_tag_metadata(self, el):
        """Extract meta data."""

        if self.type == 'odp':
            if el.namespace and el.namespace == self.namespaces['draw'] and el.name == 'page-thumbnail':
                name = el.attrs.get('draw:page-number', '')
                self.additional_context = 'slide{}:'.format(name)
        super().extract_tag_metadata(el)

    def reset(self):
        """Reset anything needed on each iteration."""

        self.type = None
        self.additional_context = ""
        super().reset()

    def get_sub_node(self, node):
        """Extract node from document if desired."""

        subnode = node.find('office:document')
        if subnode:
            mimetype = subnode.attrs['office:mimetype']
            self.type = MIMEMAP[mimetype]
            node = node.find('office:body')
        return node

    def _filter(self, text, context, encoding):
        """Filter the source text."""

        content = []
        soup = bs4.BeautifulSoup(text, self.parser)
        soup = self.get_sub_node(soup)
        blocks, attributes, comments = self.to_text(soup)
        if self.comments:
            for c, desc in comments:
                content.append(filters.SourceText(c, context + ': ' + desc, encoding, self.type + 'comment'))
        if self.attributes:
            for a, desc in attributes:
                content.append(filters.SourceText(a, context + ': ' + desc, encoding, self.type + 'attribute'))
        for b, desc in blocks:
            content.append(filters.SourceText(b, context + ': ' + desc, encoding, self.type + 'content'))
        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse XML file."""

        sources = []
        if encoding:
            with codecs.open(source_file, 'r', encoding=encoding) as f:
                src = f.read()
            sources.extend(self._filter(src, source_file, encoding))
        else:
            for content, filename, enc in self.get_content(source_file):
                sources.extend(self._filter(content, source_file, enc))
        return sources

    def sfilter(self, source):
        """Filter."""

        sources = []
        if source.text[:4].encode(source.encoding) != b'PK\x03\x04':
            sources.extend(self._filter(source.text, source.context, source.encoding))
        else:
            for content, filename, enc in self.get_content(io.BytesIO(source.text.encode(source.encoding))):
                sources.extend(self._filter(content, source.context, enc))
        return sources


def get_plugin():
    """Return the filter."""

    return OdfFilter



"""Test ODF plugin."""
from .. import util


class TestODFFilter(util.PluginTestCase):
    """Test ODF files."""

    def test_odt(self):
        """Test `odt` files."""

        config = self.dedent(
            """
            matrix:
            - name: odt
              sources:
              - 'tests/**/*.odt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.odf
            """
        ).format(self.tempdir)
        self.mktemp('.odt.yml', config, 'utf-8')
        self.assert_spellcheck('.odt.yml', ['tihs', 'smoe', 'txet'])

    def test_fodt(self):
        """Test `fodt` files."""

        config = self.dedent(
            """
            matrix:
            - name: fodt
              sources:
              - 'tests/**/*.fodt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.odf
            """
        ).format(self.tempdir)
        self.mktemp('.fodt.yml', config, 'utf-8')
        self.assert_spellcheck('.fodt.yml', ['tihs', 'smoe', 'txet'])

    def test_odp(self):
        """Test `odp` files."""

        config = self.dedent(
            """
            matrix:
            - name: odp
              sources:
              - 'tests/**/*.odp'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.odf
            """
        ).format(self.tempdir)
        self.mktemp('.odp.yml', config, 'utf-8')
        self.assert_spellcheck('.odp.yml', ['tihs', 'smoe', 'txet'])

    def test_ods(self):
        """Test `ods` files."""

        config = self.dedent(
            """
            matrix:
            - name: ods
              sources:
              - 'tests/**/*.ods'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.odf
            """
        ).format(self.tempdir)
        self.mktemp('.ods.yml', config, 'utf-8')
        self.assert_spellcheck('.ods.yml', ['tihs', 'smoe', 'txet'])

    def test_odt_chained(self):
        """Test `odt` chained files."""

        config = self.dedent(
            """
            matrix:
            - name: odt
              default_encoding: latin-1
              sources:
              - 'tests/**/*.odt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text
              - pyspelling.filters.odf
            """
        ).format(self.tempdir)
        self.mktemp('.odt.yml', config, 'utf-8')
        self.assert_spellcheck('.odt.yml', ['tihs', 'smoe', 'txet'])

    def test_fodt_chained(self):
        """Test `fodt` chained files."""

        config = self.dedent(
            """
            matrix:
            - name: fodt
              sources:
              - 'tests/**/*.fodt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text
              - pyspelling.filters.odf
            """
        ).format(self.tempdir)
        self.mktemp('.fodt.yml', config, 'utf-8')
        self.assert_spellcheck('.fodt.yml', ['tihs', 'smoe', 'txet'])



"""Test Markdown plugin."""
from .. import util


class TestMarkdown(util.PluginTestCase):
    """Test Markdown plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: markdown
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.markdown:
                  markdown_extensions:
                  - markdown.extensions.fenced_code:
              - pyspelling.filters.html:
                   ignores:
                   - code
                   - pre
            """
        ).format(self.tempdir)
        self.mktemp('.markdown.yml', config, 'utf-8')

    def test_markdown(self):
        """Test Markdown."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']

        template = self.dedent(
            """
            ## Title

            {}

            Line `slajdl alsjs`

            ```
            skjadf alsdkjls
            ```
            """
        ).format(
            '\n'.join(bad_words + good_words),
        )

        print(template)
        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.markdown.yml', bad_words)


class TestMarkdownChained(util.PluginTestCase):
    """Test chained Markdown plugin."""

    def setup_fs(self):
        """Setup file system."""

        config = self.dedent(
            """
            matrix:
            - name: markdown
              sources:
              - '{}/**/*.txt'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text:
              - pyspelling.filters.markdown:
                  markdown_extensions:
                  - markdown.extensions.fenced_code:
              - pyspelling.filters.html:
                   ignores:
                   - code
                   - pre
            """
        ).format(self.tempdir)
        self.mktemp('.markdown.yml', config, 'utf-8')

    def test_markdown_after_text(self):
        """Test Markdown."""

        bad_words = ['helo', 'begn']
        good_words = ['yes', 'word']

        template = self.dedent(
            """
            ## Title

            {}

            Line `slajdl alsjs`

            ```
            skjadf alsdkjls
            ```
            """
        ).format(
            '\n'.join(bad_words + good_words)
        )

        self.mktemp('test.txt', template, 'utf-8')
        self.assert_spellcheck('.markdown.yml', bad_words)



"""Markdown filter."""
from .. import filters
import codecs
import markdown


class MarkdownFilter(filters.Filter):
    """Spelling Python."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "markdown_extensions": []
        }

    def setup(self):
        """Setup."""

        extensions = []
        extension_configs = {}
        for item in self.config['markdown_extensions']:
            if isinstance(item, str):
                extensions.append(item)
            else:
                k, v = list(item.items())[0]
                extensions.append(k)
                if v is not None:
                    extension_configs[k] = v
        self.markdown = markdown.Markdown(extensions=extensions, extension_configs=extension_configs)

    def filter(self, source_file, encoding):  # noqa A001
        """Parse Markdown file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        return [filters.SourceText(self._filter(text), source_file, encoding, 'markdown')]

    def _filter(self, text):
        """Filter markdown."""

        self.markdown.reset()
        return self.markdown.convert(text)

    def sfilter(self, source):
        """Filter."""

        return [filters.SourceText(self._filter(source.text), source.context, source.encoding, 'markdown')]


def get_plugin():
    """Return the filter."""

    return MarkdownFilter



"""Test Office Open XML plugin."""
from .. import util


class TestOOOXMLFilter(util.PluginTestCase):
    """Test Office Open XML files."""

    def test_docx(self):
        """Test `docx` files."""

        config = self.dedent(
            """
            matrix:
            - name: docx
              sources:
              - 'tests/**/*.docx'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.ooxml
            """
        ).format(self.tempdir)
        self.mktemp('.docx.yml', config, 'utf-8')
        self.assert_spellcheck('.docx.yml', ['tihs', 'smoe', 'txet'])

    def test_pptx(self):
        """Test `pptx` files."""

        config = self.dedent(
            """
            matrix:
            - name: pptx
              sources:
              - 'tests/**/*.pptx'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.ooxml
            """
        ).format(self.tempdir)
        self.mktemp('.pptx.yml', config, 'utf-8')
        self.assert_spellcheck('.pptx.yml', ['tihs', 'smoe', 'txet'])

    def test_xlsx(self):
        """Test `xlsx` files."""

        config = self.dedent(
            """
            matrix:
            - name: xlsx
              sources:
              - 'tests/**/*.xlsx'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.ooxml
            """
        ).format(self.tempdir)
        self.mktemp('.xlsx.yml', config, 'utf-8')
        self.assert_spellcheck('.xlsx.yml', ['tihs', 'smoe', 'txet'])

    def test_docx_chained(self):
        """Test `docx` chained files."""

        config = self.dedent(
            """
            matrix:
            - name: docx
              default_encoding: latin-1
              sources:
              - 'tests/**/*.docx'
              aspell:
                lang: en
              hunspell:
                d: en_US
              pipeline:
              - pyspelling.filters.text
              - pyspelling.filters.ooxml
            """
        ).format(self.tempdir)
        self.mktemp('.docx.yml', config, 'utf-8')
        self.assert_spellcheck('.docx.yml', ['tihs', 'smoe', 'txet'])



"""Wildcard flow control."""
from .. import flow_control
from wcmatch import fnmatch


class WildcardFlowControl(flow_control.FlowControl):
    """Control flow of objects in the pipeline with wildcard patterns."""

    FNMATCH_FLAGS = fnmatch.N | fnmatch.B | fnmatch.I | fnmatch.S

    def __init__(self, options):
        """Initialization."""

        super().__init__(options)
        self.setup()

    def get_default_config(self):
        """Get default configuration."""

        return {
            'allow': ['*'],
            'halt': [],
            'skip': []
        }

    def setup(self):
        """Get default configuration."""

        self.allow = self.config['allow']
        self.halt = self.config['halt']
        self.skip = self.config['skip']

    def match(self, category, pattern):
        """Match the category."""

        return fnmatch.fnmatch(category, pattern, flags=self.FNMATCH_FLAGS)

    def adjust_flow(self, category):
        """Adjust the flow of source control objects."""

        status = flow_control.SKIP
        for allow in self.allow:
            if self.match(category, allow):
                status = flow_control.ALLOW
                for skip in self.skip:
                    if self.match(category, skip):
                        status = flow_control.SKIP
                for halt in self.halt:
                    if self.match(category, halt):
                        status = flow_control.HALT
                if status != flow_control.ALLOW:
                    break
        return status


def get_plugin():
    """Get flow controller."""

    return WildcardFlowControl



"""C++ filter."""
import re
import codecs
from .. import filters
import textwrap
import struct
import sys

TARGETS = r'''(?x)
(?P<comments>
    (?P<block>/\*[^*]*\*+(?:[^/*][^*]*\*+)*/) |                 # multi-line comments
    (?P<start>^)?(?P<leading_space>[ \t]*)?(?P<line>//(?:{})*)  # single line comments
) |
(?P<strings>
    {}
    "(?:\\.|[^"\\\n])*" |                                         # double quotes
    '(?:\\.|[^'\\\n])*'                                           # single quotes
) |
(?P<code>
    .[^/"' \t{}]*?                                               # everything else
)
'''

CPP_STRING = r'''
(?P<raw>(?:L|u8?|U)?R)"(?P<delim>[^\n ()\t]*)\(.*?\)(?P=delim)" |
(?:L|u8?|U)"(?:\\.|[^"\\])*" |
(?:L|u8?|U)'(?:\\.|[^'\\])*' |
'''

RE_CPP = re.compile(TARGETS.format(r'\\.|[^\\\n]+', CPP_STRING, 'RLuU'), re.DOTALL | re.MULTILINE)

RE_GENERIC = re.compile(TARGETS.format('[^\n]', '', ''), re.DOTALL | re.MULTILINE)

TRIGRAPHS = {
    '??=': '#',
    '??/': '\\',
    '??\'': '^',
    '??(': '[',
    '??)': ']',
    '??!': '|',
    '??<': '{',
    '??>': '}',
    '??-': '~'
}

RE_TRIGRAPHS = re.compile(r'\?{2}[=/\'()!<>-]')

BACK_SLASH_TRANSLATION = {
    "\\a": '\a',
    "\\b": '\b',
    "\\f": '\f',
    "\\r": '\r',
    "\\t": '\t',
    "\\n": '\n',
    "\\v": '\v',
    "\\\\": '\\',
    '\\"': '"',
    "\\'": "'",
    "\\?": "?",
    "\\\n": ''
}

RE_UESC = re.compile(
    r'''(?x)
    (?P<special>\\['"abfrtnv?\\\n])|
    (?P<char>\\U[\da-fA-F]{8}|\\u[\da-fA-F]{4}|\\x[\da-fA-F]{1,})|
    (?P<oct>\\[0-7]{1,3})
    '''
)

RE_ESC = re.compile(
    r'''(?x)
    (?P<special>\\['"abfrtnv?\\\n])|
    (?P<char>\\U[\da-fA-F]{8}|\\u[\da-fA-F]{4}|(?:\\x[\da-fA-F]{1,})+)|
    (?P<oct>(?:\\[0-7]{1,3})+)
    '''
)

RE_VALID_STRING_TYPES = re.compile(r'^(?:\*|(?:[rusl]\*?)+)$', re.I)

RE_ITER_STRING_TYPES = re.compile(r'(\*|[rusl]\*?)', re.I)

BIT8 = 1
BIT16 = 2
BIT32 = 4

BIG_ENDIAN = 0x10
LITTLE_ENDIAN = 0x20
CURRENT_ENDIAN = BIG_ENDIAN if sys.byteorder == 'big' else LITTLE_ENDIAN

BYTE_STORE = {
    (BIT8 | BIG_ENDIAN): '>B',
    (BIT16 | BIG_ENDIAN): '>H',
    (BIT32 | BIG_ENDIAN): '>I',
    (BIT8 | LITTLE_ENDIAN): '<B',
    (BIT16 | LITTLE_ENDIAN): '<H',
    (BIT32 | LITTLE_ENDIAN): '<I'
}


class CppFilter(filters.Filter):
    """C++ style comment filter."""

    def __init__(self, options, default_encoding='utf-8'):
        """Initialization."""

        self.pattern = RE_GENERIC
        self.trigraphs = False
        super().__init__(options, default_encoding)

    def get_default_config(self):
        """Get default configuration."""

        return {
            "block_comments": True,
            "line_comments": True,
            "group_comments": False,
            "prefix": 'cpp',
            "generic_mode": False,
            "trigraphs": False,
            "strings": False,
            "decode_escapes": True,
            "exec_charset": 'utf-8',
            "wide_exec_charset": 'utf-32',
            "charset_size": 1,
            "wide_charset_size": 4,
            "string_types": "sul"
        }

    def validate_options(self, k, v):
        """Validate options."""

        super().validate_options(k, v)
        if k == 'charset_size':
            if v not in (1, 2, 4):
                raise ValueError("{}: '{}' is an unsupported charset size".format(self.__class__.__name__, v))
        elif k == 'wide_charset_size':
            if v not in (2, 4):
                raise ValueError("{}: '{}' is an unsupported wide charset size".format(self.__class__.__name__, v))
        elif k in ('exec_charset', 'wide_exec_charset'):
            # See if parsing fails.
            self.get_encoding_name(v)
        elif k == 'string_types':
            if RE_VALID_STRING_TYPES.match(v) is None:
                raise ValueError("{}: '{}' does not define valid string types".format(self.__class__.__name__, v))

    def eval_string_type(self, text, is_string=False):
        """Evaluate string type."""

        stype = set()
        wstype = set()

        for m in RE_ITER_STRING_TYPES.finditer(text):
            value = m.group(0)
            if value == '*':
                wstype.add('l')
                wstype.add('s')
                wstype.add('u')
                wstype.add('r')
            elif value.endswith('*'):
                wstype.add(value[0].lower())
            else:
                stype.add(value.lower())
        if is_string and 'u' not in stype and 'l' not in stype:
            stype.add('s')

        return stype, wstype

    def match_string(self, stype):
        """Match string type."""

        return not (stype - self.string_types) or bool(stype & self.wild_string_types)

    def get_string_type(self, text):
        """Get string type."""

        return self.eval_string_type(text, True)[0]

    def get_encoding_name(self, name):
        """Get encoding name."""

        name = codecs.lookup(
            filters.PYTHON_ENCODING_NAMES.get(name, name).lower()
        ).name

        if name.startswith(('utf-32', 'utf-16')):
            name = name[:6]
            if CURRENT_ENDIAN == BIG_ENDIAN:
                name += '-be'
            else:
                name += '-le'

        if name == 'utf-8-sig':
            name = 'utf-8'
        return name

    def setup(self):
        """Setup."""

        self.blocks = self.config['block_comments']
        self.lines = self.config['line_comments']
        self.group_comments = self.config['group_comments']
        self.prefix = self.config['prefix']
        self.generic_mode = self.config['generic_mode']
        self.strings = self.config['strings']
        self.trigraphs = self.config['trigraphs']
        self.decode_escapes = self.config['decode_escapes']
        self.charset_size = self.config['charset_size']
        self.wide_charset_size = self.config['wide_charset_size']
        self.exec_charset = self.get_encoding_name(self.config['exec_charset'])
        self.wide_exec_charset = self.get_encoding_name(self.config['wide_exec_charset'])
        self.string_types, self.wild_string_types = self.eval_string_type(self.config['string_types'])
        if not self.generic_mode:
            self.pattern = RE_CPP

    def evaluate_block(self, groups):
        """Evaluate block comments."""

        if self.blocks:
            self.block_comments.append([groups['block'][2:-2], self.line_num, self.current_encoding])

    def evaluate_inline_tail(self, groups):
        """Evaluate inline comments at the tail of source code."""

        if self.lines:
            self.line_comments.append([groups['line'][2:].replace('\\\n', ''), self.line_num, self.current_encoding])

    def evaluate_inline(self, groups):
        """Evaluate inline comments on their own lines."""

        # Consecutive lines with only comments with same leading whitespace
        # will be captured as a single block.
        if self.lines:
            if (
                self.group_comments and
                self.line_num == self.prev_line + 1 and
                groups['leading_space'] == self.leading
            ):
                self.line_comments[-1][0] += '\n' + groups['line'][2:].replace('\\\n', '')
            else:
                self.line_comments.append(
                    [groups['line'][2:].replace('\\\n', ''), self.line_num, self.current_encoding]
                )
            self.leading = groups['leading_space']
            self.prev_line = self.line_num

    def evaluate_unicode(self, value):
        """Evaluate Unicode."""

        if value.startswith('u8'):
            length = 1
            value = value[3:-1]
            encoding = 'utf-8'
        elif value.startswith('u'):
            length = 2
            value = value[2:-1]
            encoding = 'utf-16'
        else:
            length = 4
            value = value[2:-1]
            encoding = 'utf-32'

        def replace_unicode(m):
            """Replace Unicode."""

            groups = m.groupdict()
            esc = m.group(0)
            value = esc
            if groups.get('special'):
                # Handle basic string escapes.
                value = BACK_SLASH_TRANSLATION[esc]
            elif groups.get('char') or groups.get('oct'):
                # Handle character escapes.
                integer = int(esc[2:], 16) if groups.get('char') else int(esc[1:], 8)
                if (
                    (length < 2 and integer <= 0xFF) or
                    (length < 4 and integer <= 0xFFFF) or
                    (length >= 4 and integer <= 0x10FFFF)
                ):
                    try:
                        value = chr(integer)
                    except Exception:
                        value = ' '
            return value

        return self.norm_nl(RE_UESC.sub(replace_unicode, value).replace('\x00', '\n')), encoding

    def evaluate_normal(self, value):
        """Evaluate normal string."""

        if value.startswith('L'):
            size = self.wide_charset_size
            encoding = self.wide_exec_charset
            value = value[2:-1]
            pack = BYTE_STORE[size | CURRENT_ENDIAN]
        else:
            size = self.charset_size
            encoding = self.exec_charset
            value = value[1:-1]
            pack = BYTE_STORE[size | CURRENT_ENDIAN]
        max_value = 2 ** (size * 8) - 1

        def replace(m):
            """Replace."""

            groups = m.groupdict()
            esc = m.group(0)
            value = esc
            if groups.get('special'):
                # Handle basic string escapes.
                value = BACK_SLASH_TRANSLATION[esc]
            elif groups.get('char'):
                # Handle hex/Unicode character escapes
                if value.startswith('\\x'):
                    values = [int(x, 16) for x in value[2:].split('\\x')]
                    for i, v in enumerate(values):
                        if v <= max_value:
                            values[i] = struct.pack(pack, v)
                        else:
                            values[i] = b' '
                    value = b''.join(values).decode(encoding, errors='replace')
                else:
                    integer = int(value[2:], 16)
                    value = chr(integer).encode(encoding, errors='replace').decode(encoding)
            elif groups.get('oct'):
                # Handle octal escapes.
                values = [int(x, 8) for x in value[1:].split('\\')]
                for i, v in enumerate(values):
                    if v <= max_value:
                        values[i] = struct.pack(pack, v)
                    else:
                        values[i] = b' '
                value = b''.join(values).decode(encoding, errors='replace')
            return value

        return self.norm_nl(RE_ESC.sub(replace, value)).replace('\x00', '\n'), encoding

    def evaluate_strings(self, groups):
        """Evaluate strings."""

        if self.strings:
            encoding = self.current_encoding
            if self.generic_mode:
                # Generic assumes no escapes rules.
                self.quoted_strings.append([groups['strings'][1:-1], self.line_num, encoding])
            else:
                value = groups['strings']
                stype = set()
                if value.endswith('"'):
                    stype = self.get_string_type(value[:value.index('"')].lower().replace('8', ''))
                if not self.match_string(stype) or value.endswith("'"):
                    return
                if 'r' in stype:
                    # Handle raw strings. We can handle even if decoding is disabled.
                    olen = len(groups.get('raw')) + len(groups.get('delim')) + 2
                    clen = len(groups.get('delim')) + 2
                    value = self.norm_nl(value[olen:-clen].replace('\x00', '\n'))
                elif (
                    self.decode_escapes and not value.startswith(('\'', '"')) and
                    'l' not in stype
                ):
                    # Decode Unicode string. May have added unsupported chars, so use `UTF-8`.
                    value, encoding = self.evaluate_unicode(value)
                elif self.decode_escapes:
                    # Decode normal strings.
                    value, encoding = self.evaluate_normal(value)
                else:
                    # Don't decode and just return string content.
                    value = self.norm_nl(value[value.index('"') + 1:-1]).replace('\x00', '\n')

                if value:
                    self.quoted_strings.append([value, self.line_num, encoding])

    def evaluate(self, m):
        """Search for comments."""

        g = m.groupdict()
        if g["strings"]:
            self.evaluate_strings(g)
            self.line_num += g['strings'].count('\n')
        elif g["code"]:
            self.line_num += g["code"].count('\n')
        else:
            if g['block']:
                self.evaluate_block(g)
            elif g['start'] is None:
                self.evaluate_inline_tail(g)
            else:
                self.evaluate_inline(g)
            self.line_num += g['comments'].count('\n')

    def extend_src_text(self, content, context, text_list, category):
        """Extend the source text list with the gathered text data."""

        prefix = self.prefix + '-' if self.prefix else ''

        for comment, line, encoding in text_list:
            content.append(
                filters.SourceText(
                    textwrap.dedent(comment),
                    "%s (%d)" % (context, line),
                    encoding,
                    prefix + category
                )
            )

    def extend_src(self, content, context):
        """Extend source list."""

        self.extend_src_text(content, context, self.block_comments, 'block-comment')
        self.extend_src_text(content, context, self.line_comments, 'line-comment')
        self.extend_src_text(content, context, self.quoted_strings, 'string')

    def process_trigraphs(self, m):
        """Process trigraphs."""

        return TRIGRAPHS[m.group(0)]

    def find_content(self, text):
        """Find content."""

        if self.trigraphs:
            text = RE_TRIGRAPHS.sub(self.process_trigraphs, text)

        for m in self.pattern.finditer(self.norm_nl(text)):
            self.evaluate(m)

    def _filter(self, text, context, encoding):
        """Filter JavaScript comments."""

        content = []
        self.current_encoding = encoding
        self.line_num = 1
        self.prev_line = -1
        self.leading = ''
        self.block_comments = []
        self.line_comments = []
        self.quoted_strings = []

        self.find_content(text)
        self.extend_src(content, context)

        return content

    def filter(self, source_file, encoding):  # noqa A001
        """Parse CPP file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()

        return self._filter(text, source_file, encoding)

    def sfilter(self, source):
        """Filter."""

        return self._filter(source.text, source.context, source.encoding)


def get_plugin():
    """Get filter."""

    return CppFilter
